import {
  assertEquals,
  assertRejects,
  assertSpyCall,
  assertSpyCalls,
  describe,
  it,
  returnsNext,
  Stub,
  stub,
} from './dev_deps.ts'
import {
  exportedForTesting,
  PostgresVolunteeringDB,
  TopicsSkillsVersion,
  VolunteeringDB,
  VolunteeringDBRow,
} from './volunteering_db.ts'
import { GraphQLError, postgres } from './deps.ts'
import { GeoAPIClient } from './geo_api_client.ts'
import { Engagement } from './types.ts'
import { logger, LogSeverity } from './logging.ts'

logger.setUpLogger(
  'development',
  LogSeverity.EMERGENCY,
)

// taken from https://stackoverflow.com/a/2450976 ("Fisher-Yates Shuffle")
function shuffle<A>(array: A[]): A[] {
  let currentIndex = array.length
  // While there remain elements to shuffle...
  while (currentIndex != 0) {
    // Pick a remaining element...
    const randomIndex = Math.floor(Math.random() * currentIndex)
    currentIndex-- // And swap it with the current element.
    ;[array[currentIndex], array[randomIndex]] = [
      array[randomIndex],
      array[currentIndex],
    ]
  }
  return array
}

const testCoordinates = { lat: 123, lon: 321 }
const testGeometry = {
  type: 'Polygon',
  'coordinates': [[[3.9265487, 52.576948399], [13.9304862, 52.574110199]]],
}

const withMockedDependencies = (queryObjectResult: VolunteeringDBRow[] = []) =>
(
  test: (
    vDB: VolunteeringDB,
    geoApiClient: GeoAPIClient,
  ) => unknown | Promise<unknown>,
) => {
  const mockClient = (() => Promise.resolve(queryObjectResult)) as unknown as postgres.Sql<
    Record<string | number, never>
  >
  const mockedGeoApiClient = {} as unknown as GeoAPIClient
  stub(
    mockedGeoApiClient,
    'resolveCoordinates',
    returnsNext([Promise.resolve(testCoordinates)]),
  )
  stub(
    mockedGeoApiClient,
    'resolveGeometry',
    returnsNext([Promise.resolve(testGeometry)]),
  )
  const volunteeringDB = new PostgresVolunteeringDB(
    mockClient,
    'mock-image-proxy-base-url',
    mockedGeoApiClient,
  )
  test(volunteeringDB, mockedGeoApiClient)
}

const embeddingV1Fixtures = {
  // deno-fmt-ignore
  embedding_array: [
    1,0,0,0,0,0,0,0,0,1,
    0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,1,
    1,0,0,0,0,0,0,0,0,1,
    0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,1,
  ],
}

const embeddingV2Fixtures = {
  // deno-fmt-ignore
  embedding_array_v2: [
    // topics V2
    1,0,0,0,0,0,0,0,0,1, // animal-welfare, ..., gender-equality
    0,0,0,0,0,0,0,0,0,0,
    0,0,0,1,             // youth-development
    // skills V2
    1,0,0,0,0,0,0,0,0,1, // administration, ..., event-planning
    0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,1,   // tutoring-coaching
  ],
}

const dbFixtures = {
  row1: {
    id: 1,
    title: 'moin',
    description: 'moin blabla',
    link: 'https://moin.bla',
    image: 'https://moin.bla/blub.jpg',
    source: 'moin_bla',
    uses: 'blib',
    orga: 'blob',
    imageorga: 'https://moin.bla/blob.jpg',
    location: 'Blibhausen',
    latitude: 12.3456,
    longitude: 65.4321,
    embedding_array: JSON.stringify(embeddingV1Fixtures.embedding_array),
    embedding_array_v2: JSON.stringify(embeddingV2Fixtures.embedding_array_v2),
    cosine_similarity: 0,
    distance_in_meters: 0,
    total_results: 1,
  } as VolunteeringDBRow,
}

const assertEngagementMatchesRow = (version: TopicsSkillsVersion, row: VolunteeringDBRow, engagement: Engagement) => {
  assertEquals(engagement.id, `${row.id}`)
  assertEquals(engagement.title, row.title)
  assertEquals(engagement.description, row.description)
  assertEquals(engagement.url, row.link)
  assertEquals(
    engagement.imageUrl,
    'mock-image-proxy-base-url/crop:0:0/resize:auto:1024/plain/' + encodeURIComponent(row.image!),
  )
  assertEquals(engagement.source, row.source)
  assertEquals(engagement.categories, undefined)
  assertEquals(engagement.organizer, {
    name: row.orga,
    imageUrl: row.imageorga,
  })
  assertEquals(engagement.location, row.location)
  assertEquals(engagement.latitude, row.latitude)
  assertEquals(engagement.longitude, row.longitude)
  switch (version) {
    case 'v1':
      assertEquals(engagement.topics, [
        'agriculture-food', // first item (index 0)
        'disabilities-inclusion', // tenth item (index 9)
        'water-ocean', // last item  (index 27)
      ])
      assertEquals(engagement.skills, [
        'adaptability', // first item (index 0)
        'conflict-management', // tenth item (index 9)
        'writing-translation', // last item  (index 49)
      ])
      break
    case 'v2':
      assertEquals(engagement.topicsV2, [
        'animal-welfare', // first item (index 0)
        'gender-equality', // tenth item (index 9)
        'youth-development', // last item  (index 23)
      ])
      assertEquals(engagement.skillsV2, [
        'administration', // first item (index 0)
        'event-planning', // tenth item (index 9)
        'tutoring-coaching', // last item  (index 28)
      ])
      break
  }
  assertEquals(engagement.matchedTopics, undefined)
  assertEquals(engagement.matchedSkills, undefined)
  assertEquals(engagement.matchedGeoLocationDistanceInMeters, undefined)
}

describe('VolunteeringDB', () => {
  describe('queryEmbeddingVector', () => {
    it('correctly constructs a query embedding vector', () => {
      const expectedLength = 28 /* #topics */ + 50 /* #skills */
      // deno-fmt-ignore
      const expectedVector: number[] = embeddingV1Fixtures.embedding_array
      // deno-fmt-ignore
      const actualVector = exportedForTesting.queryEmbeddingVector([
          "agriculture-food",       // first item (index 0)
          "disabilities-inclusion", // tenth item (index 9)
          "water-ocean",            // last item  (index 27)
      ], [
          "adaptability",           // first item (index 0)
          "conflict-management",    // tenth item (index 9)
          "writing-translation",    // last item  (index 49)
      ]);
      assertEquals(actualVector.length, expectedLength)
      assertEquals(actualVector, expectedVector)
    })
  })
  describe('queryEmbeddingVectorV2', () => {
    it('correctly constructs a query embedding vector', () => {
      const expectedVector: number[] = embeddingV2Fixtures.embedding_array_v2
      // deno-fmt-ignore
      const actualVector = exportedForTesting.queryEmbeddingVectorV2([
        'animal-welfare', // first item (index 0)
        'gender-equality', // tenth item (index 9)
        'youth-development', // last item  (index 23)
      ], [
        'administration', // first item (index 0)
        'event-planning', // tenth item (index 9)
        'tutoring-coaching', // last item  (index 28)
      ]);
      assertEquals(actualVector.length, expectedVector.length)
      assertEquals(actualVector, expectedVector)
    })
  })
  describe('decodeMatchesFromEmbedding', () => {
    it('correctly reverse-decodes topics and skills matches from V1 embeddings', () => {
      // deno-fmt-ignore
      const embedding: number[] = [
        0.9,0,0,0,0,0,0,0,0,0.8,
        0,0,0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0.85,
        0.85,0,0,-0.5,0,0,0,0,0,0.99,
        0,0,0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0,0,1.0,
      ]
      // ordered by confidence DESC with queried for terms first
      const expectedMatchedTopics = [
        'agriculture-food', // first item (index 0)
        'water-ocean', // last item  (index 27)
        'disabilities-inclusion', // tenth item (index 9)
      ]
      // ordered by confidence DESC with queried for terms first
      const expectedMatchedSkills = [
        'writing-translation', // last item  (index 49)
        'conflict-management', // tenth item (index 9)
        'adaptability', // first item (index 0)
      ]
      const allTopics = Array.from(
        exportedForTesting.TOPICS_VECTOR_INDICES.keys(),
      )
      const allSkills = Array.from(
        exportedForTesting.SKILLS_VECTOR_INDICES.keys(),
      )
      const matches = exportedForTesting.decodeMatchesFromEmbedding(
        embedding,
        allTopics,
        allSkills,
        'v1',
      )
      assertEquals(matches.topics, expectedMatchedTopics)
      assertEquals(matches.skills, expectedMatchedSkills)
    })
    it('correctly reverse-decodes topics and skills matches from V2 embeddings', () => {
      // deno-fmt-ignore
      const embedding: number[] = [
        // topics V2
        0.9,0,0,0,0,0,0,0,0,0.8,
        0,0,0,0,0,0,0,0,0,0,
        0,0,0,0.85,
        // skills V2
        0.85,0,0,-0.5,0,0,0,0,0,0.99,
        0,0,0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0,1.0,
      ]

      // ordered by confidence DESC with queried for terms first
      const expectedMatchedTopics = [
        'animal-welfare', // first item (index 0)
        'youth-development', // last item  (index 23)
        'gender-equality', // tenth item (index 9)
      ]
      // ordered by confidence DESC with queried for terms first
      const expectedMatchedSkills = [
        'tutoring-coaching', // last item  (index 28)
        'event-planning', // tenth item (index 9)
        'administration', // first item (index 0)
      ]
      const allTopics = Array.from(
        exportedForTesting.TOPICS_V2_VECTOR_INDICES.keys(),
      )
      const allSkills = Array.from(
        exportedForTesting.SKILLS_V2_VECTOR_INDICES.keys(),
      )
      const matches = exportedForTesting.decodeMatchesFromEmbedding(
        embedding,
        allTopics,
        allSkills,
        'v2',
      )
      assertEquals(matches.topics, expectedMatchedTopics)
      assertEquals(matches.skills, expectedMatchedSkills)
    })
    it('filters out V1 topics and skills matches with confidence below MATCH_CONFIDENCE_THRESHOLD', () => {
      // deno-fmt-ignore
      const embedding: number[] = [
        0.9,0,0,0,0,0,0,0,0,0.3,
        0,0,0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0.85,
        0.3,0,0,-0.5,0,0,0,0,0,0.99,
        0,0,0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0,0,1.0,
      ]
      // ordered by confidence DESC with queried for terms first
      const expectedMatchedTopics = [
        'agriculture-food', // first item (index 0)
        'water-ocean', // last item  (index 27)
      ]
      // ordered by confidence DESC with queried for terms first
      const expectedMatchedSkills = [
        'writing-translation', // last item  (index 49)
        'conflict-management', // tenth item (index 9)
      ]
      const allTopics = Array.from(
        exportedForTesting.TOPICS_VECTOR_INDICES.keys(),
      )
      const allSkills = Array.from(
        exportedForTesting.SKILLS_VECTOR_INDICES.keys(),
      )
      const matches = exportedForTesting.decodeMatchesFromEmbedding(
        embedding,
        allTopics,
        allSkills,
        'v1',
      )
      assertEquals(matches.topics, expectedMatchedTopics)
      assertEquals(matches.skills, expectedMatchedSkills)
    })
    it('filters out V2 topics and skills matches with confidence below MATCH_CONFIDENCE_THRESHOLD', () => {
      // deno-fmt-ignore
      const embedding: number[] = [
        // topics V2
        0.9,0,0,0,0,0,0,0,0,0.3,
        0,0,0,0,0,0,0,0,0,0,
        0,0,0,0.85,
        // skills V2
        0.3,0,0,-0.5,0,0,0,0,0,0.99,
        0,0,0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0,1.0,
      ]
      // ordered by confidence DESC with queried for terms first
      const expectedMatchedTopics = [
        'animal-welfare', // first item (index 0)
        'youth-development', // last item  (index 23)
      ]
      // ordered by confidence DESC with queried for terms first
      const expectedMatchedSkills = [
        'tutoring-coaching', // last item  (index 28)
        'event-planning', // tenth item (index 9)
      ]
      const allTopics = Array.from(
        exportedForTesting.TOPICS_V2_VECTOR_INDICES.keys(),
      )
      const allSkills = Array.from(
        exportedForTesting.SKILLS_V2_VECTOR_INDICES.keys(),
      )
      const matches = exportedForTesting.decodeMatchesFromEmbedding(
        embedding,
        allTopics,
        allSkills,
        'v2',
      )
      assertEquals(matches.topics, expectedMatchedTopics)
      assertEquals(matches.skills, expectedMatchedSkills)
    })
    it('ranks V1 topics and skills matches queried for higher', () => {
      // deno-fmt-ignore
      const embedding: number[] = [
        0.85,0,0,0,0,0,0,0,0,0.9,
        0,0,0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0.8,0.9,
        0.85,0,0,-0.5,0,0,0,0,0,0.99,
        0,0,0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0,0,1.0,
      ]
      // ordered by confidence DESC with queried for terms first
      const expectedMatchedTopics = [
        'disabilities-inclusion', // tenth item (index 9)
        'agriculture-food', // first item (index 0)
        'water-ocean', // last item (
        'urban-rural-development',
      ]
      // ordered by confidence DESC with queried for terms first
      const expectedMatchedSkills = [
        'writing-translation', // last item  (index 49)
        'conflict-management', // tenth item (index 9)
        'adaptability', // first item (index 0)
      ]
      const queriedTopics = ['agriculture-food', 'disabilities-inclusion']
      const queriedSkills = ['writing-translation']
      const matches = exportedForTesting.decodeMatchesFromEmbedding(
        embedding,
        queriedTopics,
        queriedSkills,
        'v1',
      )
      assertEquals(matches.topics, expectedMatchedTopics)
      assertEquals(matches.skills, expectedMatchedSkills)
    })
    it('ranks V2 topics and skills matches queried for higher', () => {
      // deno-fmt-ignore
      const embedding: number[] = [
        // topics V2
        0.85,0,0,0,0,0,0,0,0,0.9,
        0,0,0,0,0,0,0,0,0,0,
        0,0.81,0,0.9,
        // skills V2
        0.85,0,0,-0.5,0,0,0,0,0,0.99,
        0,0,0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0,1.0,
      ]
      // ordered by confidence DESC with queried for terms first
      const expectedMatchedTopics = [
        'gender-equality', // tenth item (index 9)
        'sports-recreation',
        'youth-development', // last item  (index 23)
        'animal-welfare', // first item (index 0)
      ]
      // ordered by confidence DESC with queried for terms first
      const expectedMatchedSkills = [
        'tutoring-coaching', // last item  (index 28)
        'event-planning', // tenth item (index 9)
        'administration', // first item (index 0)
      ]
      const queriedTopics = ['sports-recreation', 'gender-equality']
      const queriedSkills = ['tutoring-coaching']
      const matches = exportedForTesting.decodeMatchesFromEmbedding(
        embedding,
        queriedTopics,
        queriedSkills,
        'v2',
      )
      assertEquals(matches.topics, expectedMatchedTopics)
      assertEquals(matches.skills, expectedMatchedSkills)
    })
  })

  describe('sortMatches', () => {
    it('sortMatches ranks topics and skills matches by confidence descending and queried for terms higher', () => {
      const sorted = exportedForTesting.sortMatches(
        ['a', 'c'],
        shuffle([
          ['a', 0.1],
          ['b', 0.2],
          ['c', 0.3],
          ['d', 0.4],
          ['e', 0.5],
        ]),
      )
      assertEquals(sorted, [['c', 0.3], ['a', 0.1], ['e', 0.5], ['d', 0.4], [
        'b',
        0.2,
      ]])
    })
  })

  describe('querying engagement by id', () => {
    it('correctly parses engagement', () => {
      withMockedDependencies([dbFixtures.row1])(async (volunteeringDB) => {
        const response = await volunteeringDB.engagement({ id: '1' })
        assertEngagementMatchesRow('v1', dbFixtures.row1, response!)
        assertEngagementMatchesRow('v2', dbFixtures.row1, response!)
      })
    })

    it('throws NOT_FOUND error for empty result', () => {
      withMockedDependencies([])(async (volunteeringDB) => {
        const query = () => volunteeringDB.engagement({ id: '1' })
        await assertRejects(query, GraphQLError, 'Not found')
      })
    })
  })
  describe('findRecos', () => {
    it('wraps search results with matchedTopics, matchedSkills and matchedGeoLocationDistanceInMeters', () => {
      withMockedDependencies([dbFixtures.row1])(async (volunteeringDB, _geoApiClient) => {
        const response = await volunteeringDB.engagementRecos('v1', {
          limit: 10,
          offset: 0,
          topics: ['disabilities-inclusion', 'water-ocean'],
          skills: ['adaptability', 'writing-translation'],
          geolocationId: 'mock-geolocation-id',
        })
        assertEquals(response.totalResults, 1)
        assertEquals(response.data.length, 1)

        const reco = response.data[0]
        assertEngagementMatchesRow('v1', dbFixtures.row1, reco.engagement)

        assertEquals(reco.matchedTopics, [
          'disabilities-inclusion', // tenth item (index 9)
          'water-ocean', // last item  (index 27)
          'agriculture-food', // first item (index 0)
        ])
        assertEquals(reco.matchedSkills, [
          'adaptability', // first item (index 0)
          'writing-translation', // last item  (index 49)
          'conflict-management', // tenth item (index 9)
        ])
        assertEquals(reco.matchedGeoLocationDistanceInMeters, 0)
      })
    })
    it('wraps search results with matchedTopicsV2, matchedSkillsV2 and matchedGeoLocationDistanceInMeters', () => {
      withMockedDependencies([dbFixtures.row1])(async (volunteeringDB, _geoApiClient) => {
        const response = await volunteeringDB.engagementRecos('v2', {
          limit: 10,
          offset: 0,
          topics: ['animal-welfare', 'youth-development'],
          skills: ['event-planning', 'tutoring-coaching'],
          geolocationId: 'mock-geolocation-id',
        })
        assertEquals(response.totalResults, 1)
        assertEquals(response.data.length, 1)

        const reco = response.data[0]
        assertEngagementMatchesRow('v2', dbFixtures.row1, reco.engagement)

        assertEquals(reco.matchedTopicsV2, [
          'animal-welfare',
          'youth-development',
          'gender-equality',
        ])
        assertEquals(reco.matchedSkillsV2, [
          'event-planning',
          'tutoring-coaching',
          'administration',
        ])
        assertEquals(reco.matchedGeoLocationDistanceInMeters, 0)
      })
    })
    it('resolves latitude/longitude using the GeoAPI client when geolocationId is given', () => {
      withMockedDependencies()(
        (volunteeringDB, geoAPIClient) => {
          volunteeringDB.engagementRecos('v2', {
            limit: 10,
            offset: 0,
            topics: [],
            skills: [],
            geolocationId: 'mock-geolocation-id',
          })
          assertSpyCall(geoAPIClient.resolveCoordinates as Stub, 0, {
            args: ['mock-geolocation-id'],
          })
        },
      )
    })
    it('does not resolve latitude/longitude when geolocationId is not given', () => {
      withMockedDependencies()(
        (volunteeringDB, geoAPIClient) => {
          volunteeringDB.engagementRecos('v2', {
            limit: 10,
            offset: 0,
            topics: ['agriculture-food'],
            skills: [],
          })
          assertSpyCalls(geoAPIClient.resolveCoordinates as Stub, 0)
        },
      )
    })
    it('falls back to recommend based only on topics and skills if given and geo location resolution fails', () => {
      withMockedDependencies()(
        async (volunteeringDB, geoAPIClient) => {
          stub(
            geoAPIClient,
            'resolveCoordinates',
            returnsNext([Promise.reject('boom!')]),
          )

          const result = await volunteeringDB.engagementRecos('v2', {
            limit: 10,
            offset: 0,
            topics: ['agriculture-food'],
            skills: [],
            geolocationId: 'mock-geolocation-id',
          })
          assertEquals(result.totalResults, 0)
          assertEquals(result.data, [])
        },
      )
    })
    it('fails with error message if no topics and skills are given and geo location resolution fails', () => {
      withMockedDependencies()(
        async (volunteeringDB, geoAPIClient) => {
          stub(
            geoAPIClient,
            'resolveCoordinates',
            returnsNext([Promise.reject('boom!')]),
          )
          await assertRejects(
            () =>
              volunteeringDB.engagementRecos('v2', {
                limit: 10,
                offset: 0,
                topics: [],
                skills: [],
                geolocationId: 'mock-geolocation-id',
              }),
            Error,
            'geo location resolution failed and no topics or skills given',
          )
        },
      )
    })
  })

  describe('filterEngagements', () => {
    it('resolves geolocation geometry using the GeoAPI client when geolocationId is given', () => {
      withMockedDependencies()(
        (volunteeringDB, geoAPIClient) => {
          volunteeringDB.filteredEngagements('v2', {
            limit: 10,
            offset: 0,
            topics: [],
            skills: [],
            geolocationId: 'mock-geolocation-id',
          })
          assertSpyCall(geoAPIClient.resolveGeometry as Stub, 0, {
            args: ['mock-geolocation-id'],
          })
        },
      )
    })
    it('does not resolve geolocation geometry when geolocationId is not given', () => {
      withMockedDependencies()(
        (volunteeringDB, geoAPIClient) => {
          volunteeringDB.filteredEngagements('v2', {
            limit: 10,
            offset: 0,
            topics: ['agriculture-food'],
            skills: [],
          })
          assertSpyCalls(geoAPIClient.resolveGeometry as Stub, 0)
        },
      )
    })
    it('fails with error message if geolocation resolution fails', () => {
      withMockedDependencies()(
        async (volunteeringDB, geoAPIClient) => {
          stub(
            geoAPIClient,
            'resolveGeometry',
            returnsNext([Promise.reject('boom!')]),
          )
          await assertRejects(
            () =>
              volunteeringDB.filteredEngagements('v2', {
                limit: 10,
                offset: 0,
                topics: ['agriculture-food'],
                skills: [],
                geolocationId: 'mock-geolocation-id',
              }),
            Error,
            'Geolocation resolution failed',
          )
        },
      )
    })
    it('fails with error message if no topics, skills and location are given', () => {
      withMockedDependencies()(
        async (volunteeringDB, _geoAPIClient) => {
          await assertRejects(
            () =>
              volunteeringDB.filteredEngagements('v2', {
                limit: 10,
                offset: 0,
                topics: [],
                skills: [],
              }),
            Error,
            'At least one topic, skill or location is required',
          )
        },
      )
    })
  })
})
